/*
 * kgraft_patch_bsc1031481
 *
 * Fix for CVE-2017-7294, bsc#1031481
 *
 *  SLE12-SP2 commit:
 *  560498d4a9a33ccaf4016993435a543f49501508
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <drm/drmP.h>
#include <drm/vmwgfx_drm.h>
#include <drm/ttm/ttm_memory.h>

#include "local.h"

#define KGR_PATCHED_MODULE      "vmwgfx"

static uint64_t *kgr_vmw_user_surface_size;
static const struct svga3d_surface_desc (*kgr_svga3d_surface_descs)[];
static int (*kgr_vmw_resource_init)(struct vmw_private *dev_priv, struct vmw_resource *res, bool delay_id, void (*res_free) (struct vmw_resource *res), const struct vmw_res_func *func);
static const struct vmw_res_func *kgr_vmw_legacy_surface_func;
static void (*kgr_vmw_resource_activate)(struct vmw_resource *res, void (*hw_destroy) (struct vmw_resource *));
static void (*kgr_vmw_hw_surface_destroy)(struct vmw_resource *res);
static void (*kgr_vmw_user_surface_free)(struct vmw_resource *res);
static struct vmw_resource *(*kgr_vmw_resource_reference)(struct vmw_resource *res);
static void (*kgr_vmw_user_surface_base_release)(struct ttm_base_object **p_base);
static void (*kgr_vmw_resource_unreference)(struct vmw_resource **p_res);
static int (*kgr_ttm_read_lock)(struct ttm_lock *lock, bool interruptible);
static void (*kgr_ttm_read_unlock)(struct ttm_lock *lock);
static int (*kgr_ttm_base_object_init)(struct ttm_object_file *tfile, struct ttm_base_object *base, bool shareable, enum ttm_object_type object_type, void (*refcount_release) (struct ttm_base_object **), void (*ref_obj_release) (struct ttm_base_object *, enum ttm_ref_type ref_type));
static void (*kgr_ttm_mem_global_free)(struct ttm_mem_global *glob, uint64_t amount);
static int (*kgr_ttm_mem_global_alloc)(struct ttm_mem_global *glob, uint64_t memory, bool no_wait, bool interruptible);
static size_t (*kgr_ttm_round_pot)(size_t size);
static void (*kgr_drm_err)(const char *format, ...);
static const struct vmw_res_func *kgr_vmw_gb_surface_func;
static int (*kgr_vmw_user_dmabuf_alloc)(struct vmw_private *dev_priv, struct ttm_object_file *tfile, uint32_t size, bool shareable, uint32_t *handle, struct vmw_dma_buffer **p_dma_buf, struct ttm_base_object **p_base);
static int (*kgr_ttm_prime_object_init)(struct ttm_object_file *tfile, size_t size, struct ttm_prime_object *prime, bool shareable, enum ttm_object_type type, void (*refcount_release) (struct ttm_base_object **), void (*ref_obj_release) (struct ttm_base_object *, enum ttm_ref_type ref_type));
static struct drm_master *(*kgr_drm_master_get)(struct drm_master *master);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "vmw_user_surface_size", (void *)&kgr_vmw_user_surface_size },
	{ "svga3d_surface_descs", (void *)&kgr_svga3d_surface_descs },
	{ "vmw_resource_init", (void *)&kgr_vmw_resource_init },
	{ "vmw_legacy_surface_func", (void *)&kgr_vmw_legacy_surface_func },
	{ "vmw_resource_activate", (void *)&kgr_vmw_resource_activate },
	{ "vmw_hw_surface_destroy", (void *)&kgr_vmw_hw_surface_destroy },
	{ "vmw_user_surface_free", (void *)&kgr_vmw_user_surface_free },
	{ "vmw_resource_reference", (void *)&kgr_vmw_resource_reference },
	{ "vmw_user_surface_base_release", (void *)&kgr_vmw_user_surface_base_release },
	{ "vmw_resource_unreference", (void *)&kgr_vmw_resource_unreference },
	{ "ttm_read_lock", (void *)&kgr_ttm_read_lock },
	{ "ttm_read_unlock", (void *)&kgr_ttm_read_unlock },
	{ "ttm_base_object_init", (void *)&kgr_ttm_base_object_init },
	{ "ttm_mem_global_free", (void *)&kgr_ttm_mem_global_free },
	{ "ttm_mem_global_alloc", (void *)&kgr_ttm_mem_global_alloc },
	{ "ttm_round_pot", (void *)&kgr_ttm_round_pot },
	{ "drm_err", (void *)&kgr_drm_err },
	{ "vmw_gb_surface_func", (void *)&kgr_vmw_gb_surface_func },
	{ "vmw_user_dmabuf_alloc", (void *)&kgr_vmw_user_dmabuf_alloc },
	{ "ttm_prime_object_init", (void *)&kgr_ttm_prime_object_init },
	{ "drm_master_get", (void *)&kgr_drm_master_get },
};

/* inlined */
static inline struct vmw_private *kgr_vmw_priv(struct drm_device *dev)
{
	return (struct vmw_private *)dev->dev_private;
}

/* inlined */
static inline struct vmw_fpriv *kgr_vmw_fpriv(struct drm_file *file_priv)
{
	return (struct vmw_fpriv *)file_priv->driver_priv;
}

/* inlined */
static inline struct vmw_master *kgr_vmw_master(struct drm_master *master)
{
	return (struct vmw_master *) master->driver_priv;
}

/* inlined */
static inline const struct svga3d_surface_desc *kgr_svga3dsurface_get_desc(SVGA3dSurfaceFormat format)
{
	/*
	 * kgraft hack
	 *  -if (format < ARRAY_SIZE(*kgr_svga3d_surface_descs))
	 */
	if (format < 145)
		return &(*kgr_svga3d_surface_descs)[format];

	return &(*kgr_svga3d_surface_descs)[SVGA3D_FORMAT_INVALID];
}

/* inlined */
static inline struct ttm_mem_global *kgr_vmw_mem_glob(struct vmw_private *dev_priv)
{
	return (struct ttm_mem_global *) dev_priv->mem_global_ref.object;
}

/* inlined */
static inline void kgr_svga3dsurface_get_size_in_blocks(const struct svga3d_surface_desc *desc,
	const surf_size_struct *pixel_size, surf_size_struct *block_size)
{
	block_size->width = DIV_ROUND_UP(pixel_size->width,
					 desc->block_size.width);
	block_size->height = DIV_ROUND_UP(pixel_size->height,
					  desc->block_size.height);
	block_size->depth = DIV_ROUND_UP(pixel_size->depth,
					 desc->block_size.depth);
}

/* inlined */
static inline u32 kgr_svga3dsurface_calculate_pitch(const struct svga3d_surface_desc *desc,
	const surf_size_struct *size)
{
	u32 pitch;
	surf_size_struct blocks;

	kgr_svga3dsurface_get_size_in_blocks(desc, size, &blocks);

	pitch = blocks.width * desc->pitch_bytes_per_block;

	return pitch;
}

/* inlined */
static inline bool kgr_svga3dsurface_is_planar_surface(const struct svga3d_surface_desc *desc)
{
	return (desc->block_desc & SVGA3DBLOCKDESC_PLANAR_YUV) != 0;
}

/* inlined */
static inline u32 kgr_clamped_umul32(u32 a, u32 b)
{
	uint64_t tmp = (uint64_t) a*b;
	return (tmp > (uint64_t) ((u32) -1)) ? (u32) -1 : tmp;
}

/* inlined */
static inline u32 kgr_svga3dsurface_get_image_buffer_size(const struct svga3d_surface_desc *desc,
	const surf_size_struct *size, u32 pitch)
{
	surf_size_struct image_blocks;
	u32 slice_size, total_size;

	kgr_svga3dsurface_get_size_in_blocks(desc, size, &image_blocks);

	if (kgr_svga3dsurface_is_planar_surface(desc)) {
		total_size = kgr_clamped_umul32(image_blocks.width,
					    image_blocks.height);
		total_size = kgr_clamped_umul32(total_size, image_blocks.depth);
		total_size = kgr_clamped_umul32(total_size, desc->bytes_per_block);
		return total_size;
	}

	if (pitch == 0)
		pitch = kgr_svga3dsurface_calculate_pitch(desc, size);

	slice_size = kgr_clamped_umul32(image_blocks.height, pitch);
	total_size = kgr_clamped_umul32(slice_size, image_blocks.depth);

	return total_size;
}

/* inlined */
static inline void kgr_vmw_fifo_resource_inc(struct vmw_private *dev_priv)
{
	atomic_inc(&dev_priv->num_fifo_resources);
}

/* inlined */
static inline void kgr_vmw_fifo_resource_dec(struct vmw_private *dev_priv)
{
	atomic_dec(&dev_priv->num_fifo_resources);
}

/* optimized */
static int kgr_vmw_surface_init(struct vmw_private *dev_priv,
	struct vmw_surface *srf, void (*res_free)(struct vmw_resource *res))
{
	int ret;
	struct vmw_resource *res = &srf->res;

	BUG_ON(res_free == NULL);
	if (!dev_priv->has_mob)
		kgr_vmw_fifo_resource_inc(dev_priv);
	ret = kgr_vmw_resource_init(dev_priv, res, true, res_free,
				(dev_priv->has_mob) ? kgr_vmw_gb_surface_func :
				kgr_vmw_legacy_surface_func);

	if (unlikely(ret != 0)) {
		if (!dev_priv->has_mob)
			kgr_vmw_fifo_resource_dec(dev_priv);
		res_free(res);
		return ret;
	}

	/*
	 * The surface won't be visible to hardware until a
	 * surface validate.
	 */

	INIT_LIST_HEAD(&srf->view_list);
	kgr_vmw_resource_activate(res, kgr_vmw_hw_surface_destroy);
	return ret;
}

/* patched function */
int kgr_vmw_surface_define_ioctl(struct drm_device *dev, void *data,
	struct drm_file *file_priv)
{
	struct vmw_private *dev_priv = kgr_vmw_priv(dev);
	struct vmw_user_surface *user_srf;
	struct vmw_surface *srf;
	struct vmw_resource *res;
	struct vmw_resource *tmp;
	union drm_vmw_surface_create_arg *arg =
	    (union drm_vmw_surface_create_arg *)data;
	struct drm_vmw_surface_create_req *req = &arg->req;
	struct drm_vmw_surface_arg *rep = &arg->rep;
	struct ttm_object_file *tfile = kgr_vmw_fpriv(file_priv)->tfile;
	struct drm_vmw_size __user *user_sizes;
	int ret;
	int i, j;
	uint32_t cur_bo_offset;
	struct drm_vmw_size *cur_size;
	struct vmw_surface_offset *cur_offset;
	uint32_t num_sizes;
	uint32_t size;
	const struct svga3d_surface_desc *desc;

	if (unlikely(*kgr_vmw_user_surface_size == 0))
		*kgr_vmw_user_surface_size = kgr_ttm_round_pot(sizeof(*user_srf)) +
			128;

	num_sizes = 0;
	/*
	 * Fix CVE-2017-7294
	 *  -for (i = 0; i < DRM_VMW_MAX_SURFACE_FACES; ++i)
	 *  +4 lines
	 */
	for (i = 0; i < DRM_VMW_MAX_SURFACE_FACES; ++i) {
		if (req->mip_levels[i] > DRM_VMW_MAX_MIP_LEVELS)
			return -EINVAL;
		num_sizes += req->mip_levels[i];
	}

	if (num_sizes > DRM_VMW_MAX_SURFACE_FACES *
	    DRM_VMW_MAX_MIP_LEVELS)
		return -EINVAL;

	size = *kgr_vmw_user_surface_size + 128 +
		kgr_ttm_round_pot(num_sizes * sizeof(struct drm_vmw_size)) +
		kgr_ttm_round_pot(num_sizes * sizeof(struct vmw_surface_offset));


	desc = kgr_svga3dsurface_get_desc(req->format);
	if (unlikely(desc->block_desc == SVGA3DBLOCKDESC_NONE)) {
		KGR_DRM_ERROR("Invalid surface format for surface creation.\n");
		KGR_DRM_ERROR("Format requested is: %d\n", req->format);
		return -EINVAL;
	}

	ret = kgr_ttm_read_lock(&dev_priv->reservation_sem, true);
	if (unlikely(ret != 0))
		return ret;

	ret = kgr_ttm_mem_global_alloc(kgr_vmw_mem_glob(dev_priv),
				   size, false, true);
	if (unlikely(ret != 0)) {
		if (ret != -ERESTARTSYS)
			KGR_DRM_ERROR("Out of graphics memory for surface"
				  " creation.\n");
		goto out_unlock;
	}

	user_srf = kzalloc(sizeof(*user_srf), GFP_KERNEL);
	if (unlikely(user_srf == NULL)) {
		ret = -ENOMEM;
		goto out_no_user_srf;
	}

	srf = &user_srf->srf;
	res = &srf->res;

	srf->flags = req->flags;
	srf->format = req->format;
	srf->scanout = req->scanout;

	memcpy(srf->mip_levels, req->mip_levels, sizeof(srf->mip_levels));
	srf->num_sizes = num_sizes;
	user_srf->size = size;

	srf->sizes = kmalloc(srf->num_sizes * sizeof(*srf->sizes), GFP_KERNEL);
	if (unlikely(srf->sizes == NULL)) {
		ret = -ENOMEM;
		goto out_no_sizes;
	}
	srf->offsets = kmalloc(srf->num_sizes * sizeof(*srf->offsets),
			       GFP_KERNEL);
	if (unlikely(srf->sizes == NULL)) {
		ret = -ENOMEM;
		goto out_no_offsets;
	}

	user_sizes = (struct drm_vmw_size __user *)(unsigned long)
	    req->size_addr;

	ret = copy_from_user(srf->sizes, user_sizes,
			     srf->num_sizes * sizeof(*srf->sizes));
	if (unlikely(ret != 0)) {
		ret = -EFAULT;
		goto out_no_copy;
	}

	srf->base_size = *srf->sizes;
	srf->autogen_filter = SVGA3D_TEX_FILTER_NONE;
	srf->multisample_count = 0;

	cur_bo_offset = 0;
	cur_offset = srf->offsets;
	cur_size = srf->sizes;

	for (i = 0; i < DRM_VMW_MAX_SURFACE_FACES; ++i) {
		for (j = 0; j < srf->mip_levels[i]; ++j) {
			uint32_t stride = kgr_svga3dsurface_calculate_pitch
				(desc, cur_size);

			cur_offset->face = i;
			cur_offset->mip = j;
			cur_offset->bo_offset = cur_bo_offset;
			cur_bo_offset += kgr_svga3dsurface_get_image_buffer_size
				(desc, cur_size, stride);
			++cur_offset;
			++cur_size;
		}
	}
	res->backup_size = cur_bo_offset;
	if (srf->scanout &&
	    srf->num_sizes == 1 &&
	    srf->sizes[0].width == 64 &&
	    srf->sizes[0].height == 64 &&
	    srf->format == SVGA3D_A8R8G8B8) {

		srf->snooper.image = kmalloc(64 * 64 * 4, GFP_KERNEL);
		/* clear the image */
		if (srf->snooper.image) {
			memset(srf->snooper.image, 0x00, 64 * 64 * 4);
		} else {
			KGR_DRM_ERROR("Failed to allocate cursor_image\n");
			ret = -ENOMEM;
			goto out_no_copy;
		}
	} else {
		srf->snooper.image = NULL;
	}
	srf->snooper.crtc = NULL;

	user_srf->prime.base.shareable = false;
	user_srf->prime.base.tfile = NULL;
	if (drm_is_primary_client(file_priv))
		user_srf->master = kgr_drm_master_get(file_priv->master);

	/**
	 * From this point, the generic resource management functions
	 * destroy the object on failure.
	 */

	ret = kgr_vmw_surface_init(dev_priv, srf, kgr_vmw_user_surface_free);
	if (unlikely(ret != 0))
		goto out_unlock;

	/*
	 * A gb-aware client referencing a shared surface will
	 * expect a backup buffer to be present.
	 */
	if (dev_priv->has_mob && req->shareable) {
		uint32_t backup_handle;

		ret = kgr_vmw_user_dmabuf_alloc(dev_priv, tfile,
					    res->backup_size,
					    true,
					    &backup_handle,
					    &res->backup,
					    &user_srf->backup_base);
		if (unlikely(ret != 0)) {
			kgr_vmw_resource_unreference(&res);
			goto out_unlock;
		}
	}

	tmp = kgr_vmw_resource_reference(&srf->res);
	ret = kgr_ttm_prime_object_init(tfile, res->backup_size, &user_srf->prime,
				    req->shareable, VMW_RES_SURFACE,
				    kgr_vmw_user_surface_base_release, NULL);

	if (unlikely(ret != 0)) {
		kgr_vmw_resource_unreference(&tmp);
		kgr_vmw_resource_unreference(&res);
		goto out_unlock;
	}

	rep->sid = user_srf->prime.base.hash.key;
	kgr_vmw_resource_unreference(&res);

	kgr_ttm_read_unlock(&dev_priv->reservation_sem);
	return 0;
out_no_copy:
	kfree(srf->offsets);
out_no_offsets:
	kfree(srf->sizes);
out_no_sizes:
	ttm_prime_object_kfree(user_srf, prime);
out_no_user_srf:
	kgr_ttm_mem_global_free(kgr_vmw_mem_glob(dev_priv), size);
out_unlock:
	kgr_ttm_read_unlock(&dev_priv->reservation_sem);
	return ret;
}

static int kgr_patch_bsc1031481_kallsyms(void)
{
        unsigned long addr;
        int i;

        for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
                /* mod_find_symname would be nice, but it is not exported */
                addr = kallsyms_lookup_name(kgr_funcs[i].name);
                if (!addr) {
                        pr_err("kgraft-patch: symbol %s not resolved\n",
                                kgr_funcs[i].name);
                        return -ENOENT;
                }

                *(kgr_funcs[i].addr) = (void *)addr;
        }

        return 0;
}

static int kgr_patch_bsc1031481_module_notify(struct notifier_block *nb,
	unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1031481_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1031481_module_nb = {
	.notifier_call = kgr_patch_bsc1031481_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1031481_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1031481_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1031481_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1031481_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1031481_module_nb);
}
